﻿using System.Net;
using System.Net.Http;
using System.Threading.Tasks;
using FluentAssertions;
using Microsoft.Extensions.Caching.Memory;
using Microsoft.Extensions.Logging;
using NSubstitute;
using Robware.Lib.ApiClient;
using Website.Data;
using Website.Models.Auth;
using Xunit;

namespace Website.Tests.Data {
	public class AuthenticationProviderTests {
		[Fact]
		public async Task Authenticate_WithSuccessfulLoginRequest_ReturnsUser() {
			const string json = @"{""username"":""username"",""password"":""password""}";
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Post)
							 .WithUrl("/user/authenticate")
							 .WithPostBody(@"{""Username"":""username"",""Password"":""password"",""ReturnUrl"":null}")
							 .WithResponse(json)
							 .Build();

			var cache = Substitute.For<IMemoryCache>();
			var logger = Substitute.For<ILogger<AuthenticationProvider>>();

			var request = new LoginRequest {
				Username = "username",
				Password = "password"
			};

			var expectedUser = new User {
				Username = "username",
				Password = "password"
			};

			var provider = new AuthenticationProvider(httpClient, cache, new CacheDurations(), logger);
			(await provider.Authenticate(request)).Should().BeEquivalentTo(expectedUser);
		}

		[Fact]
		public async Task Authenticate_WithFailedLoginRequest_ReturnsNull() {
			const string json = @"{""username"":""username"",""password"":""password""}";
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Post)
							 .WithUrl("/user/authenticate")
							 .WithPostBody(@"{""Username"":""username"",""Password"":""password"",""ReturnUrl"":null}")
							 .WithErrorStatus(HttpStatusCode.Unauthorized)
							 .WithResponse(json)
							 .Build();

			var cache = Substitute.For<IMemoryCache>();
			var logger = Substitute.For<ILogger<AuthenticationProvider>>();

			var request = new LoginRequest {
				Username = "username",
				Password = "wrong"
			};

			var provider = new AuthenticationProvider(httpClient, cache, new CacheDurations(), logger);
			(await provider.Authenticate(request)).Should().BeNull();
		}

		[Fact]
		public async Task UpdateUserPassword_WithSuccessfulRequest_ReturnsTrue() {
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Post)
							 .WithUrl("/user/updateuserpassword")
							 .WithPostBody(@"{""Username"":""username"",""OldPassword"":""correct"",""NewPassword"":""new""}")
							 .WithResponseCode(HttpStatusCode.NoContent)
							 .Build();

			var cache = Substitute.For<IMemoryCache>();
			var logger = Substitute.For<ILogger<AuthenticationProvider>>();

			var request = new UpdatePasswordRequest {
				Username = "username",
				OldPassword = "correct",
				NewPassword = "new"
			};

			var provider = new AuthenticationProvider(httpClient, cache, new CacheDurations(), logger);
			(await provider.UpdateUserPassword(request)).Should().BeTrue();
		}

		[Fact]
		public async Task UpdateUserPassword_WithUnsuccessfulRequest_ReturnsFalse() {
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Post)
							 .WithUrl("/user/updateuserpassword")
							 .WithPostBody(@"{""Username"":""username"",""OldPassword"":""incorrect"",""NewPassword"":""new""}")
							 .WithResponseCode(HttpStatusCode.BadRequest)
							 .Build();

			var cache = Substitute.For<IMemoryCache>();
			var logger = Substitute.For<ILogger<AuthenticationProvider>>();

			var request = new UpdatePasswordRequest {
				Username = "username",
				OldPassword = "incorrect",
				NewPassword = "new"
			};

			var provider = new AuthenticationProvider(httpClient, cache, new CacheDurations(), logger);
			(await provider.UpdateUserPassword(request)).Should().BeTrue();
		}
	}
}